import hashlib
import json
import os
import os.path
import shutil
import tempfile
import zipfile
from collections import defaultdict
from datetime import datetime as dt
from functools import partial
from glob import iglob

import fastavro
from ckan.lib import uploader
from ckan.plugins import toolkit
from splitgill.search import rebuild_data

from ckanext.versioned_datastore.lib import common
from ckanext.versioned_datastore.lib.downloads.utils import (
    calculate_field_counts,
    filter_data_fields,
    get_fields,
    get_schema,
)
from ckanext.versioned_datastore.lib.query.utils import get_resources_and_versions
from ckanext.versioned_datastore.lib.utils import (
    get_database,
    idownload_implementations,
)
from ckanext.versioned_datastore.logic.download.arg_objects import (
    DerivativeArgs,
    NotifierArgs,
    QueryArgs,
    ServerArgs,
)
from ckanext.versioned_datastore.model.downloads import (
    CoreFileRecord,
    DerivativeFileRecord,
    DownloadRequest,
)

from .loaders import (
    get_derivative_generator,
    get_file_server,
    get_notifier,
    get_transformation,
)


class DownloadRunManager:
    download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
    if download_dir is None:
        raise Exception('ckanext.versioned_datastore.download_dir must be set.')
    core_dir = os.path.join(download_dir, 'core')
    custom_dir = toolkit.config.get(
        'ckanext.versioned_datastore.custom_dir', os.path.join(download_dir, 'custom')
    )

    def __init__(
        self,
        query_args: QueryArgs,
        derivative_args: DerivativeArgs,
        server_args: ServerArgs,
        notifier_args: NotifierArgs,
    ):
        # allow plugins to make changes to the args
        for plugin in idownload_implementations():
            (
                query_args,
                derivative_args,
                server_args,
                notifier_args,
            ) = plugin.download_before_init(
                query_args, derivative_args, server_args, notifier_args
            )

        if derivative_args.format == 'raw':
            # if we're serving raw files then we're not doing any transformations
            derivative_args.separate_files = True
            derivative_args.ignore_empty_fields = False
            derivative_args.transform = {}
            # we also can't use a query
            query_args.query = {}
            query_args.slug_or_doi = None

        self.query = query_args.to_schema_query()
        self.allow_non_datastore = (
            derivative_args.format == 'raw'
            and derivative_args.format_args.get('allow_non_datastore', False)
        )
        self.resource_ids_and_versions = get_resources_and_versions(
            self.query,
            allow_non_datastore=self.allow_non_datastore,
        )
        self.derivative_options = derivative_args
        # todo: i don't think this is necessary cause BaseArgs handles defaults?
        self.derivative_options.ensure_defaults_are_set()
        self.server = get_file_server(
            server_args.type,
            filename=server_args.custom_filename,
            **server_args.type_args,
        )

        # initialise core and derivative records
        self.core_record, self.derivative_record = self.check_for_records()

        # initialises a log entry in the database
        self.request = DownloadRequest(
            core_id=self.core_record.id,
            derivative_id=self.derivative_record.id,
            server_args={f: getattr(server_args, f) for f in server_args.fields},
        )
        self.request.save()

        self.notifier = get_notifier(
            notifier_args.type, request=self.request, **notifier_args.type_args
        )

        self._temp = []

        for plugin in idownload_implementations():
            plugin.download_after_init(self.request)

    @property
    def derivative_hash(self):
        """
        Unique hash for the derivative file options.
        """
        file_options = {
            f: getattr(self.derivative_options, f)
            for f in self.derivative_options.fields
        }
        file_options_hash = hashlib.sha1(json.dumps(file_options).encode('utf-8'))
        return file_options_hash.hexdigest()

    @property
    def hash(self):
        """
        Unique hash for this download request, identified by the query and the
        derivative options.
        """
        to_hash = [self.record_hash, self.derivative_hash]
        download_hash = hashlib.sha1('|'.join(to_hash).encode('utf-8'))
        return download_hash.hexdigest()

    @property
    def resource_hash(self) -> str:
        resources = sorted(self.resource_ids_and_versions.items())
        return hashlib.sha1('|'.join(map(str, resources)).encode('utf-8')).hexdigest()

    @property
    def record_hash(self):
        to_hash = [self.query.hash, self.resource_hash]
        download_hash = hashlib.sha1('|'.join(to_hash).encode('utf-8'))
        return download_hash.hexdigest()

    @property
    def core_folder_path(self):
        """
        Location of the core files for this query.
        """
        return os.path.join(self.core_dir, self.query.hash)

    def run(self):
        """
        Run the download process.
        """
        try:
            # refresh the db objects because they were probably retrieved in a
            # different session (the __init__ is run by the main ckan process,
            # this is run by the download worker)
            self.request = DownloadRequest.get(self.request.id)
            self.core_record = CoreFileRecord.get(self.core_record.id)
            self.derivative_record = DerivativeFileRecord.get(self.derivative_record.id)

            self.notifier.notify_start()
            self.request.update_status(DownloadRequest.state_initial)

            # generate core file if needed
            # technically we don't need core files if the format is 'raw', but we'll
            # generate them anyway for consistency/avoidance of thousands of ifs, and
            # also so we have some field metadata
            self.generate_core()

            # generate derivative file if needed
            self.generate_derivative()

            # finish up
            self.request.update_status(DownloadRequest.state_complete)
            url = self.server.serve(self.request)
            self.notifier.notify_end(url)
        except Exception as e:
            current_state = self.request.state
            self.request.update_status(
                DownloadRequest.state_failed,
                f'{current_state}, {e.__class__.__name__}: {str(e)}',
            )
            self.notifier.notify_error()
            raise e
        finally:
            for t in self._temp:
                try:
                    shutil.rmtree(t)
                except FileNotFoundError:
                    pass
            for plugin in idownload_implementations():
                plugin.download_after_run(self.request)

    def check_for_records(self):
        """
        Check if relevant files and records already exist and returns the records if
        they exist.

        :returns: tuple of core_record (or None) and derivative_record (or None)
        """
        # check the download dir exists
        if not os.path.exists(self.download_dir):
            os.mkdir(self.download_dir)

        # also check the core dir exists
        if not os.path.exists(self.core_dir):
            os.mkdir(self.core_dir)

        # initialise empty variables
        core_record = None
        derivative_record = None

        # search for derivative first
        fp = os.path.join(self.download_dir, f'{self.hash}.zip')
        if os.path.exists(fp):
            existing_file = fp
        else:
            # check old-style names with the request id in front as well
            fn = f'*_{self.hash}.zip'
            existing_file = next(iglob(os.path.join(self.download_dir, fn)), None)
        if existing_file is not None:
            # could return multiple options, sorted by most recent first
            possible_records = DerivativeFileRecord.get_by_filepath(existing_file)
            if possible_records:
                # use the most recent one
                derivative_record = possible_records[0]
                core_record = derivative_record.core_record

        # if the core record hasn't been found by searching for the derivative, try and
        # find it now
        if core_record is None:
            if os.path.exists(self.core_folder_path):
                # could return multiple options, sorted by most recent first
                possible_records = CoreFileRecord.get_by_hash(
                    self.query.hash, self.resource_hash
                )
                if possible_records:
                    # use the most recent one
                    core_record = possible_records[0]

        if core_record is None:
            core_record = CoreFileRecord(
                query_hash=self.query.hash,
                query=self.query.query,
                query_version=self.query.query_version,
                resource_ids_and_versions=self.resource_ids_and_versions,
                resource_hash=self.resource_hash,
            )
            core_record.save()

        if derivative_record is None:
            derivative_record = DerivativeFileRecord(
                core_id=core_record.id,
                download_hash=self.hash,
                format=self.derivative_options.format,
                options={
                    f: getattr(self.derivative_options, f)
                    for f in self.derivative_options.fields
                    if f != 'format'
                },
            )
            derivative_record.save()

        return core_record, derivative_record

    def generate_core(self):
        """
        Generates and loads core files.

        This method will add new resources to existing core records with identical
        filters, reducing data duplication.
        :returns: the core record
        """
        if not os.path.exists(self.core_folder_path):
            os.makedirs(self.core_folder_path)

        # check if there are new resources to generate
        existing_files = os.listdir(self.core_folder_path)
        resources_to_generate = {
            rid: v
            for rid, v in self.core_record.resource_ids_and_versions.items()
            if f'{rid}_{v}.avro' not in existing_files
            and v != common.NON_DATASTORE_VERSION
        }

        # todo: are these needed?
        resource_totals = {k: 0 for k in self.core_record.resource_ids_and_versions}
        field_counts = {k: None for k in self.core_record.resource_ids_and_versions}

        # get info for resources that weren't just generated first, so we know if they
        # need regenerating
        existing_resources = [
            k
            for k, v in self.core_record.resource_ids_and_versions.items()
            if k not in resources_to_generate and v != common.NON_DATASTORE_VERSION
        ]
        for resource_id in existing_resources:
            # find a matching core record
            record = CoreFileRecord.find_resource(
                self.query.hash,
                resource_id,
                self.core_record.resource_ids_and_versions[resource_id],
            )
            if record:
                resource_totals[resource_id] = record.resource_totals[resource_id]
                field_counts[resource_id] = record.field_counts[resource_id]
            else:
                # if there's no record we should regenerate the avro file
                resource_version = self.core_record.resource_ids_and_versions[
                    resource_id
                ]
                core_file_path = os.path.join(
                    self.core_folder_path, f'{resource_id}_{resource_version}.avro'
                )
                os.remove(core_file_path)
                resources_to_generate[resource_id] = resource_version

        if len(resources_to_generate) > 0:
            for resource_id, version in resources_to_generate.items():
                self.request.update_status(DownloadRequest.state_core_gen, resource_id)

                resource_totals[resource_id] = 0
                field_counts[resource_id] = calculate_field_counts(
                    resource_id, version, self.query
                )

                database = get_database(resource_id)
                search = database.search(version).filter(self.query.to_dsl())

                schema = fastavro.parse_schema(
                    get_schema(resource_id, version, self.query)
                )

                fn = f'{resource_id}_{version}.avro'
                fp = os.path.join(self.core_folder_path, fn)

                codec_kwargs = dict(codec='bzip2', codec_compression_level=9)
                chunk_size = 10000
                with open(fp, 'wb') as f:
                    fastavro.writer(f, schema, [], **codec_kwargs)

                def _flush(record_block):
                    with open(fp, 'a+b') as outfile:
                        fastavro.writer(outfile, None, record_block, **codec_kwargs)

                chunk = []
                for hit in search.scan():
                    data = rebuild_data(hit.data.to_dict())
                    resource_totals[resource_id] += 1
                    chunk.append(data)
                    if len(chunk) == chunk_size:
                        _flush(chunk)
                        chunk = []
                _flush(chunk)

        non_datastore_resources = [
            k
            for k, v in self.resource_ids_and_versions.items()
            if v == common.NON_DATASTORE_VERSION
        ]
        for resource_id in non_datastore_resources:
            resource_totals[resource_id] = 1
            field_counts[resource_id] = {}

        self.core_record.update(
            resource_totals=resource_totals,
            total=sum(filter(None, resource_totals.values())),
            field_counts=field_counts,
        )

        return self.core_record

    def generate_derivative(self):
        """
        Generates derivative files, if necessary.

        :returns:
        """

        self.request.update_status(DownloadRequest.state_derivative_gen)

        if self.derivative_record.filepath is None:
            # if this _is_ defined, we don't need to generate the file
            zip_name = f'{self.hash}.zip'
            zip_path = os.path.join(self.download_dir, zip_name)

            self.derivative_record.update(filepath=zip_path)

            # for keeping track of elapsed generation time
            start = dt.utcnow()
            # for storing build files
            temp_dir = tempfile.mkdtemp()
            self._temp.append(temp_dir)

            manifest = {
                'download_id': self.request.id,
                'resources': {},
                'separate_files': self.derivative_options.separate_files,
                'file_format': self.derivative_options.format,
                'format_args': self.derivative_options.format_args,
                'ignore_empty_fields': self.derivative_options.ignore_empty_fields,
                'transform': self.derivative_options.transform,
                'total_records': self.core_record.total,
                'start': start.isoformat(),
                # these get filled in later but we'll initialise them here
                'files': [],
                'end': None,
                'duration_in_seconds': 0,
            }

            if self.derivative_options.format != 'raw':
                # set up the derivative generators; each generator creates one
                # component. components = individual file groups within the main zip,
                # e.g. one CSV for each resource (multiple components), or multiple
                # files comprising a single DarwinCore archive (single component)
                fields = partial(
                    get_fields,
                    self.core_record.field_counts,
                    self.derivative_options.ignore_empty_fields,
                )
                if self.derivative_options.separate_files:
                    components = {
                        rid: get_derivative_generator(
                            self.derivative_options.format,
                            output_dir=temp_dir,
                            fields=fields(resource_id=rid),
                            resource_id=rid,
                            query=self.query,
                            **self.derivative_options.format_args,
                        )
                        for rid in self.resource_ids_and_versions
                    }
                else:
                    gen = get_derivative_generator(
                        self.derivative_options.format,
                        output_dir=temp_dir,
                        fields=fields(),
                        query=self.query,
                        **self.derivative_options.format_args,
                    )
                    components = defaultdict(lambda: gen)

                # load transformation functions
                transformations = [
                    get_transformation(t, **targs)
                    for t, targs in (self.derivative_options.transform or {}).items()
                ]

                for resource_id, version in self.resource_ids_and_versions.items():
                    # add the resource ID as the message for use in the status page
                    self.request.update_status(
                        DownloadRequest.state_derivative_gen, resource_id
                    )
                    if (
                        len(self.resource_ids_and_versions) > 1
                        and self.core_record.resource_totals[resource_id] == 0
                    ):
                        # don't generate empty files unless there's only one resource
                        continue
                    derivative_generator = components[resource_id]
                    core_file_path = os.path.join(
                        self.core_folder_path, f'{resource_id}_{version}.avro'
                    )
                    with derivative_generator, open(core_file_path, 'rb') as core_file:
                        for record in fastavro.reader(core_file):
                            # apply the transformations first
                            for transform in transformations:
                                record = transform(record)
                            # filter out fields that are empty for all records
                            if self.derivative_options.ignore_empty_fields:
                                record = filter_data_fields(
                                    record, self.core_record.field_counts[resource_id]
                                )
                            # then write the record
                            derivative_generator.write(record)

                if self.derivative_options.separate_files:
                    for generator in components.values():
                        generator.cleanup()
                else:
                    components.default_factory().cleanup()

            else:
                resource_show = toolkit.get_action('resource_show')
                for resource_id, version in self.resource_ids_and_versions.items():
                    self.request.update_status(
                        DownloadRequest.state_derivative_gen, resource_id
                    )
                    res = resource_show({}, {'id': resource_id})
                    if res.get('url_type') == 'upload':
                        upload = uploader.get_resource_uploader(res)
                        if not upload.storage_path:
                            # if ckan.storage_path is not set
                            raise Exception('Non-datastore uploads are not configured.')
                        filepath = upload.get_path(res['id'])
                        original_filename = res.get('url', '').split('/')[-1]
                        if res.get('download_original_filenames', False):
                            filename = original_filename
                        else:
                            fileext = (
                                os.path.splitext(original_filename)[-1].strip('.')
                                or res['format'].lower()
                            )
                            filename = os.path.extsep.join([resource_id, fileext])
                        shutil.copy2(filepath, os.path.join(temp_dir, filename))
                    else:
                        raise NotImplementedError(
                            f'No raw file is available for resource {resource_id}'
                        )

            # generation finished
            self.request.update_status(DownloadRequest.state_packaging)

            end = dt.utcnow()
            manifest['end'] = end.isoformat()

            duration = (end - start).total_seconds()
            manifest['duration_in_seconds'] = duration

            files_to_zip = os.listdir(temp_dir) + ['manifest.json']
            manifest['files'] = files_to_zip

            # allow plugins to make changes
            for plugin in idownload_implementations():
                manifest = plugin.download_modify_manifest(manifest, self.request)

            # write out manifest
            with open(
                os.path.join(temp_dir, 'manifest.json'), 'w', encoding='utf8'
            ) as f:
                json.dump(manifest, f, sort_keys=True, indent=2, ensure_ascii=False)

            # zip everything up
            with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED, True) as z:
                for filename in files_to_zip:
                    z.write(os.path.join(temp_dir, filename), arcname=filename)

        # auth checks should have already been done, so this should have been removed
        # if not allowed
        if self.server.filename:
            # ensure the custom dir exists
            if not os.path.exists(self.custom_dir):
                os.mkdir(self.custom_dir)
            symlink_path = os.path.join(self.custom_dir, f'{self.server.filename}.zip')
            if (
                os.path.islink(symlink_path)
                and os.path.realpath(symlink_path) != self.derivative_record.filepath
            ):
                # this should only be being used by admins, so we assume it's okay to
                # overwrite old symlinks
                os.remove(symlink_path)
            # test it again and create if necessary
            if not os.path.islink(symlink_path):
                os.symlink(self.derivative_record.filepath, symlink_path)

        return self.derivative_record



from ckantools.decorators import auth


@auth(anon=True)
def datastore_search(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_basic_query(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_basic_count(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_basic_autocomplete(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_query_extent(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_basic_extent(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}



import hashlib
import secrets
import string
from contextlib import closing
from pathlib import Path
from typing import Iterable

import requests
from ckan.plugins import toolkit
from splitgill.model import Record

from ckanext.versioned_datastore.model.stats import ImportStats


def download_resource_data(resource: dict, into: Path, api_key: str) -> str:
    """
    Downloads the file specified in the resource's url field to the given path and
    returns the SHA1 hash of it. If the url is an upload url (i.e. the URL of a file
    which is stored on this CKAN instance) then the API key will be used to ensure we
    have access. This allows private datasets to have resources ingested in the
    datastore before they are made public.

    :param resource: the resource dict
    :param into: the path to the file where the data should be put
    :param api_key: the user's API key
    :returns: the hash of the downloaded file
    """
    hasher = hashlib.sha1()
    # grab the resource's data via this URL
    url = toolkit.url_for(
        'resource.download',
        id=resource['package_id'],
        resource_id=resource['id'],
        qualified=True,
    )
    # include the auth header regardless of whether the resource URL is of a file hosted
    # by this CKAN instance, or another website. We need to do this as the
    # resource.download route is protected by auth. This is safe and won't leak creds
    # because if the resource's URL is another website, the resource.download route will
    # respond with a redirect to the other site and requests won't copy those headers on
    # to the request to the redirected URL
    headers = {'Authorization': api_key}
    with closing(requests.get(url, stream=True, headers=headers)) as r:
        r.raise_for_status()
        with into.open('wb') as f:
            for chunk in r.iter_content(chunk_size=8192, decode_unicode=False):
                if chunk:
                    f.write(chunk)
                    hasher.update(chunk)

    return hasher.hexdigest()


def iter_records(data: Iterable[dict], stats: ImportStats) -> Iterable[Record]:
    """
    Iterate over the dicts in the given data iterable, converting each to a Record
    object for Splitgill to ingest. The stats object will be updated periodically during
    the operation to show progress (by updating the count value).

    For each dict in the data stream, the _id key is checked to see if it exists. If it
    does exist, the associated value is used as the record ID for the record created
    from that dict. If it does not exist, a new _id value is added to the dict and used
    as the new record's ID.

    New record IDs are generated sequentially to maintain insertion order of these
    records within this data stream. Unless there are more than 1 billion records in the
    stream, the resulting generated IDs will always be 12 characters long. If there are
    more than 1 billion records in the stream, the resulting generated IDs *may* be
    longer than 12 characters.

    IDs take the form of a 3-letter prefix concatenated with the sum of a constant value
    and the record's index in the stream (i.e. the first record is at position 0, the
    10th record is at position 9 etc) in hex. The hex representation is padded with 0s
    to ensure it is at least 9 characters long, hence achieving a 12 character total ID
    length. The constant value is a random number between 0 (inclusive) and 3294967296
    (exclusive) which is chosen because it is 1 billion less than the maximum integer
    expressible in 9 hex characters. This is where the 1 billion soft-limit on IDs of
    length 12 comes from. If the constant is chosen as 3294967295 and a billion records
    IDs are generated, the hex representation of the constant + the last index in the
    stream will be 10 characters long, resulting in a 13 character ID. Is this
    overcomplicated? Perhaps.

    :param data: an iterable of dicts representing record data
    :param stats: the current ImportStats object in use which will be updated with
        progress every 5000 records handled
    :returns: yields Record objects
    """
    stats.update(count=0)

    # choose a random 3 letter prefix for all records in this stream
    prefix = ''.join(secrets.choice(string.ascii_lowercase) for _ in range(3))
    constant = secrets.randbelow(3294967296)
    index = 1

    for record_data in data:
        if index % 5000 == 0:
            stats.update(count=index)

        record_id = record_data.get('_id')
        if not record_id:
            # generate a new id for the record
            record_id = f'{prefix}{constant + index:09x}'
        else:
            # make sure it's a str
            record_id = str(record_id)

        # put the record ID in the record's data, this is necessary if the _id wasn't
        # already in there, and it ensures it is a str if it was
        record_data['_id'] = record_id
        yield Record(record_id, record_data)
        index += 1

    stats.update(count=index)



import pytest
from mock import MagicMock
from splitgill.model import Record

from ckanext.versioned_datastore.lib.common import ALL_FORMATS, DATASTORE_ONLY_RESOURCE
from ckanext.versioned_datastore.lib.importing.options import (
    create_default_options_builder,
)
from ckanext.versioned_datastore.lib.utils import (
    get_database,
    is_datastore_only_resource,
    is_datastore_resource,
    is_ingestible,
)


@pytest.mark.usefixtures('with_vds')
def test_is_datastore_resource():
    assert not is_datastore_resource('eggs')

    # manually add some data to the database
    database = get_database('eggs')
    database.update_options(create_default_options_builder().build(), commit=False)
    database.ingest([Record.new({'beans': 'yes please'})], commit=True)
    database.sync()
    # there should be data now so this should be recognised as a datastore resource
    assert is_datastore_resource('eggs')


def test_is_datastore_only_resource():
    for yes in [
        DATASTORE_ONLY_RESOURCE,
        f'http://{DATASTORE_ONLY_RESOURCE}',
        f'https://{DATASTORE_ONLY_RESOURCE}',
    ]:
        assert is_datastore_only_resource(yes)

    for no in [
        f'ftp://{DATASTORE_ONLY_RESOURCE}',
        'this is datastore only',
        None,
        f'{DATASTORE_ONLY_RESOURCE}/{DATASTORE_ONLY_RESOURCE}',
        f'https://{DATASTORE_ONLY_RESOURCE}/nope',
    ]:
        assert not is_datastore_only_resource(no)


def test_is_ingestible():
    # all formats should be ingestible (even in uppercase)
    for fmt in ALL_FORMATS:
        assert is_ingestible({'format': fmt, 'url': MagicMock()})
        assert is_ingestible({'format': fmt.upper(), 'url': MagicMock()})
    # zip should be ingestible (even in uppercase)
    assert is_ingestible({'format': 'ZIP', 'url': MagicMock()})
    assert is_ingestible({'format': 'zip', 'url': MagicMock()})
    # a datastore only resource should be ingestible
    assert is_ingestible({'format': None, 'url': DATASTORE_ONLY_RESOURCE})

    # if there's no url then the resource is not ingestible
    assert not is_ingestible({'url': None})
    assert not is_ingestible({'format': 'csv', 'url': None})
    # if there's no format and the resource is not datastore only then it is not ingestible
    assert not is_ingestible({'format': None, 'url': 'http://banana.com/test.csv'})



import dataclasses
from typing import Any, Dict, List, Optional

from ckan.plugins import toolkit
from elasticsearch_dsl import AttrDict, MultiSearch, Search
from elasticsearch_dsl.aggs import A
from elasticsearch_dsl.query import Bool
from elasticsearch_dsl.response import Response
from splitgill.indexing.fields import DocumentField
from splitgill.search import rebuild_data, version_query

from ckanext.versioned_datastore.lib.query.search.query import Query
from ckanext.versioned_datastore.lib.query.search.sort import Sort
from ckanext.versioned_datastore.lib.utils import (
    es_client,
    get_database,
    ivds_implementations,
    unprefix_index_name,
)


@dataclasses.dataclass
class SearchRequest:
    """
    Class representing a search request.
    """

    query: Query
    size: Optional[int] = 100
    offset: Optional[int] = 0
    after: Optional[Any] = None
    sorts: List[Sort] = dataclasses.field(default_factory=list)
    fields: List[str] = dataclasses.field(default_factory=list)
    aggs: Dict[str, A] = dataclasses.field(default_factory=dict)
    # any additional filters which shouldn't be represented as part of the main query
    extra_filter: Bool = dataclasses.field(default_factory=Bool)
    # setting this to True will cause no versions to be taken into account and all data
    # will be returned. As an example of where this is useful, consider aggregations
    # across all versions of a record or resource
    force_no_version: bool = False
    # if this request has been created based on an action call then it's useful to store
    # a reference to the original data dict that the action was called with. Defaults to
    # None, indicating there is no associated (or relevant) data dict available.
    data_dict: Optional[dict] = None
    ignore_auth: bool = False
    # optional additional request parameters to be included when performing the search
    req_params: dict = dataclasses.field(default_factory=dict)

    def add_param(self, param: str, value: Any):
        """
        Add a request parameter to be passed as part of this search request. The search
        request is completed using the _msearch endpoint on Elasticsearch so only
        parameters that work on that endpoint should be added.

        :param param: the param name
        :param value: the param value
        """
        self.req_params[param] = value

    def add_sort(self, field: str, ascending: bool = True):
        """
        Convenience wrapper to add a sort to the sort list on the given field with the
        given direction.

        :param field: the field to sort on
        :param ascending: whether to sort in ascending (True, default) or descending
            (False)
        """
        self.sorts.append(Sort(field, ascending))

    def add_agg(self, name, agg_type, *args, **kwargs):
        """
        Creates an aggregation and adds it to the dict of aggregations on this request.

        :param name: the name of the aggregation
        :param agg_type: the aggregation type
        :param args: the aggregation arguments
        :param kwargs: the aggregation kwarguments
        :returns:
        """
        self.aggs[name] = A(agg_type, *args, **kwargs)

    def indexes(self) -> List[str]:
        """
        A list of the indexes this request will search over. This list is created from
        the resource_ids specified in the query.

        :returns: a list of index names, this could include wildcards
        """
        databases = map(get_database, self.query.resource_ids)

        if self.force_no_version or self.query.version is not None:
            # todo: could check latest version for each and optimise to latest?
            return [database.indices.wildcard for database in databases]
        else:
            return [database.indices.latest for database in databases]

    def set_no_results(self):
        """
        Sets the request size to zero and removes all pagination values.

        This is useful for aggregation only results or counts where you don't need any
        hits.
        """
        self.size = 0
        self.offset = 0
        self.after = None

    def get_safe_size(self) -> int:
        """
        The size that is actually going to be used in the request to Elasticsearch. This
        method returns a value based on the `self.size` property but capped between 0
        and 1000, with a default value of 100 if `self.size` is set to None.

        :returns: a number between 0 and 1000
        """
        return max(0, min(100 if self.size is None else self.size, 1000))

    def to_search(self) -> Search:
        """
        Builds an Elasticsearch Search object with the query, indexes, sorts, size, and
        aggregations set.

        :returns: a new Elasticsearch Search object
        """
        search = (
            Search()
            .index(self.indexes())
            # we want to provide an accurate count, and damn the expense
            .extra(track_total_hits=True)
        )

        search = search.query(self.query.to_dsl())

        # add the version filter, if needed
        if not self.force_no_version and self.query.version is not None:
            search = search.filter(version_query(self.query.version))

        # add any extra filters if there are any
        if self.extra_filter.to_dict()['bool']:
            search = search.filter(self.extra_filter)

        size = self.get_safe_size()
        if size == 0:
            # an empty request
            search = search.extra(size=0)
        else:
            # add one to the size so that we can work out when there are no more hits
            size += 1
            search = search.extra(size=size)

            # only offset or after should be used
            if self.offset:
                # if from + size > 10000, running the search will fail. we still want
                # to return an error instead of silently failing, but we can return a
                # better error.
                search_window = size + self.offset
                if search_window > 10000:
                    raise toolkit.ValidationError(
                        f'Search window size out of bounds: {search_window} (from: '
                        f'{self.offset}, size: {size})'
                    )
                search = search.extra(from_=self.offset)
            elif self.after is not None:
                search = search.extra(search_after=self.after)

            # set the source if required
            if self.fields:
                fields = [f'{DocumentField.DATA}.{field}' for field in self.fields]
                # add the fields we need for the ResultRecord wrapper class to work
                fields.extend([DocumentField.ID, DocumentField.VERSION])
                search = search.source(fields)

            # add sorting
            if self.sorts:
                sorts = [sort.to_sort() for sort in self.sorts]
            else:
                sorts = [{DocumentField.VERSION: 'desc'}]
            # TODO: can we use _doc here? Is it safe across indexes?
            # always add the default sort to ensure search after values are unique
            sorts.extend([Sort.desc('_id').to_sort(), {'_index': 'desc'}])
            search = search.sort(*sorts)

        # add any aggregations
        for agg_name, agg in self.aggs.items():
            search.aggs[agg_name] = agg

        return search

    def run(self) -> 'SearchResponse':
        """
        Builds the search, runs it, and returns a SearchResponse object.

        :returns: a SearchResponse object
        """
        for plugin in ivds_implementations():
            plugin.vds_before_search(self)

        search = self.to_search()

        # use a multisearch to wrap the search to avoid any issues with URL length. When
        # you query a lot of indexes you can get errors because the URL contains all the
        # index names, comma separated, and it can cause a URL to be created which is
        # beyond the allowed length of Elasticsearch. Using a multisearch solves this
        # problem because the multisearch is sent to Elasticsearch as a POST request
        # with all parts of the search, including the indexes, as part of the payload
        multi_search = MultiSearch(using=es_client()).add(search)
        multi_search = multi_search.params(**self.req_params)
        result = next(iter(multi_search.execute()))
        return SearchResponse(self, result)


@dataclasses.dataclass
class ResultRecord:
    """
    A wrapper on a hit from the response.

    Unless source fields have been used in the search request, this will be wrapping a
    version of a record.
    """

    hit: AttrDict

    @property
    def id(self) -> str:
        """
        :returns: the record's ID
        """
        return self.hit[DocumentField.ID]

    @property
    def version(self) -> int:
        """
        :returns: the record's version
        """
        return self.hit[DocumentField.VERSION]

    @property
    def data(self) -> dict:
        """
        Rebuilds the record data from the hit in the raw Elasticsearch response using
        Splitgill's rebuild_data function and returns the data as a dict.

        :returns: the record data
        """
        return rebuild_data(self.hit[DocumentField.DATA].to_dict())

    @property
    def index(self) -> str:
        """
        :returns: the index this record comes from
        """
        return self.hit.meta.index

    @property
    def resource_id(self) -> str:
        """
        :returns: the resource this record comes from
        """
        return unprefix_index_name(self.index)


@dataclasses.dataclass
class SearchResponse:
    """
    Class wrapping the Elasticsearch search response object.
    """

    request: SearchRequest
    response: Response

    def __post_init__(self):
        # cache the safe size just in case the request gets fiddled with after the
        # search is completed (it's unlikely, but it would be a real pig of a bug to
        # track down)
        self._request_size = self.request.get_safe_size()

    @property
    def count(self) -> int:
        """
        :returns: the number of documents which matched the search request
        """
        return self.response.hits.total.value

    @property
    def hits(self) -> List[ResultRecord]:
        """
        :returns: returns the list of ResultRecords which were returned in the request
        """
        hits = [ResultRecord(hit) for hit in self.response.hits]
        if len(self.response.hits) > self._request_size:
            # there are more results than requested due to our addition of +1 to the
            # size during the request, so trim off the last hit as it wasn't asked for
            return hits[:-1]
        else:
            return hits

    @property
    def data(self) -> List[dict]:
        """
        :returns: a list of just the data dicts for each hit
        """
        return [hit.data for hit in self.hits]

    @property
    def aggs(self) -> dict:
        """
        :returns: the aggregation results, keyed by the names specified in the request
        """
        return self.response.aggs.to_dict()

    @property
    def next_after(self) -> Optional[list]:
        """
        Returns the after value to be used to get the next set of results. If there are
        no more results to get, None is returned.

        :returns: None or a list
        """
        if self.count == 0:
            return None

        if len(self.response.hits) > self._request_size:
            if 'sort' in self.hits[-1].hit.meta:
                return list(self.hits[-1].hit.meta['sort'])

        return None



import re
from math import isfinite
from typing import Iterable, Optional, Union

from ckan.model import Resource, Session
from ckan.plugins import toolkit

from ckanext.versioned_datastore.lib.utils import (
    get_public_resources,
    is_datastore_resource,
)

boolean_validator = toolkit.get_validator('boolean_validator')
ignore_missing = toolkit.get_validator('ignore_missing')
int_validator = toolkit.get_validator('int_validator')
not_missing = toolkit.get_validator('not_missing')
not_empty = toolkit.get_validator('not_empty')
resource_id_exists = toolkit.get_validator('resource_id_exists')
email_validator = toolkit.get_validator('email_validator')
one_of = toolkit.get_validator('one_of')


def _populate_context(context=None):
    """
    Populates the context with a user if it doesn't already exist.

    :param context: an existing context object
    :returns: dict with at least a 'user' key, even if the value is None
    """
    context = context.copy() if context is not None else {}
    if context.get('user') is None:
        try:
            context['user'] = toolkit.g.get('user')
        except RuntimeError:
            # during e.g. testing we don't have access to toolkit.g
            context['user'] = None
    return context


def float_validator(value, context) -> float:
    """
    Checks if the value can be parsed as a float and returns it if it can, otherwise
    raises Invalid. Rejects NaN and inf.

    :param value: the value
    :param context: the context
    :returns: a float
    """
    try:
        value = float(value)
        if isfinite(value):
            return value
    except (ValueError, TypeError):
        pass
    raise toolkit.Invalid('Invalid float value')


def url_safe(value, context):
    """
    Checks if the value is safe to be included in a URL as a slug.

    :param value: the value to check
    :param context: the context in which to check
    """
    if not re.match('^[A-Za-z0-9-_]+$', value):
        raise toolkit.Invalid(
            'Only a-z, 0-9, hyphens (-) and underscores (_) are valid characters'
        )
    else:
        return value


def check_resource_id(resource_id: str, context: Optional[dict] = None) -> bool:
    context = _populate_context(context)

    # check it exists
    if not Session.query(Resource).get(resource_id):
        return False
    # check we have access to it
    try:
        toolkit.check_access('resource_show', context, {'id': resource_id})
        return True
    except toolkit.NotAuthorized:
        return False


def check_datastore_resource_id(
    resource_id: str, context: Optional[dict] = None
) -> bool:
    return check_resource_id(resource_id, context) and is_datastore_resource(
        resource_id
    )


def _deduplicate(values: Iterable[str]) -> Iterable[str]:
    """
    Simple util function to remove duplicate entries in an iterable.

    :param values: the iterable to deduplicate
    :returns: the deduplicated values as another iterable
    """
    seen = set()
    for value in values:
        if value in seen:
            continue
        seen.add(value)
        yield value


def validate_resource_ids(value: Union[str, list], context: Optional[dict] = None):
    if isinstance(value, str):
        value = value.split(',') if value else []
    if not isinstance(value, list):
        raise toolkit.Invalid('Invalid list of resource ID strings')

    context = _populate_context(context)
    public_resources = get_public_resources()
    valid_resource_ids = []
    for resource_id in _deduplicate(value):
        datastore_status = public_resources.get(resource_id)
        if datastore_status is None:
            if check_resource_id(resource_id, context):
                valid_resource_ids.append(resource_id)
        else:
            valid_resource_ids.append(resource_id)

    if value and not valid_resource_ids:
        # the user passed some resources, but none of them were datastore resources
        raise toolkit.Invalid('No resource IDs are available')

    return valid_resource_ids


def validate_datastore_resource_ids(
    value: Union[str, list], context: Optional[dict] = None
):
    if isinstance(value, str):
        value = value.split(',') if value else []
    if not isinstance(value, list):
        raise toolkit.Invalid('Invalid list of resource ID strings')

    context = _populate_context(context)
    public_resources = get_public_resources()
    valid_resource_ids = []
    for resource_id in _deduplicate(value):
        datastore_status = public_resources.get(resource_id)
        if datastore_status is None:
            datastore_status = check_datastore_resource_id(resource_id, context)
        if datastore_status:
            valid_resource_ids.append(resource_id)

    if value and not valid_resource_ids:
        # the user passed some resources, but none of them were datastore resources
        raise toolkit.Invalid('No resource IDs are datastore resources')

    return valid_resource_ids


def validate_datastore_resource_id(value: str, context: Optional[dict] = None):
    return validate_datastore_resource_ids([value], context)[0]



from typing import List, Optional

import jsonschema
from ckan.plugins import plugin_loaded, toolkit
from ckantools.decorators import action

from ckanext.versioned_datastore.lib.query.schema import (
    InvalidQuerySchemaVersionError,
    get_latest_query_version,
)
from ckanext.versioned_datastore.lib.query.search.query import SchemaQuery
from ckanext.versioned_datastore.lib.query.slugs.slugs import (
    create_nav_slug,
    create_slug,
    resolve_slug,
)
from ckanext.versioned_datastore.lib.query.utils import convert_to_multisearch
from ckanext.versioned_datastore.logic.slug import helptext, schema
from ckanext.versioned_datastore.logic.validators import validate_datastore_resource_ids


@action(schema.vds_slug_create(), helptext.vds_slug_create)
def vds_slug_create(
    resource_ids: List[str] = None,
    version: Optional[int] = None,
    query: Optional[dict] = None,
    query_version: Optional[str] = None,
    pretty_slug: bool = True,
    nav_slug: bool = False,
):
    """
    Creates a slug for the given query options. If the slug already exists then this
    will be indicated in the is_new key of the response dict.

    :param resource_ids: the resource IDs to search over (no value or an empty list will
        result in a search over all resources at the time of recovering the slug and
        searching)
    :param version: the version to search at
    :param query: the query to search with
    :param query_version: the query schema version to use
    :param pretty_slug: whether to produce a pretty slug (es muy bonita)
    :param nav_slug: whether to produce a nav type slug
    :returns: details about the slug
    """
    if query_version and query_version.lower().startswith('v0'):
        # this is an old/basic query, so we need to convert it first
        query = convert_to_multisearch(query)
        query_version = None

    schema_query = SchemaQuery(resource_ids, version, query, query_version)
    try:
        # check the query is valid before we go any further
        schema_query.validate()
    except jsonschema.ValidationError as e:
        raise toolkit.ValidationError(e.message)

    try:
        if nav_slug:
            is_new, slug = create_nav_slug(schema_query)
        else:
            is_new, slug = create_slug(schema_query, pretty_slug=pretty_slug)
    except (jsonschema.ValidationError, InvalidQuerySchemaVersionError) as e:
        raise toolkit.ValidationError(e.message)

    if slug is None:
        raise toolkit.ValidationError('Failed to generate new slug')

    return {
        'slug': slug.get_slug_string(),
        'is_new': is_new,
        'is_reserved': (
            False if nav_slug else slug.reserved_pretty_slug == slug.get_slug_string()
        ),
    }


@action(schema.vds_slug_resolve(), helptext.vds_slug_resolve, get=True)
def vds_slug_resolve(slug: str):
    """
    Resolves the given slug and returns the query details. If the provided slug string
    is a DOI and the query_dois plugin is loaded, this action also checks that plugin
    for DOIs and returns the query details of the DOI.

    :param slug: the slug to resolve
    :returns: the slug query details, or a ValidationError if no slug could be resolved
    """
    saved_search_type = None
    result = {
        'query': {},
        'query_version': '',
        'version': None,
        'resource_ids': [],
        'created': '',
        'warnings': [],
    }
    # try resolving the slug first
    resolved = resolve_slug(slug)
    if resolved:
        saved_search_type = 'slug'
        result.update(
            {
                'query': resolved.query,
                'query_version': resolved.query_version,
                'version': resolved.version,
                # todo: should we turn an empty resource_ids list into the list of all
                #       available resources at this point or let it ride until query time?
                'resource_ids': resolved.resource_ids,
                'created': resolved.created.isoformat(),
            }
        )

    # then check if it's a query DOI
    elif plugin_loaded('query_dois'):
        from ckan import model

        from ckanext.query_dois.model import QueryDOI

        resolved = model.Session.query(QueryDOI).filter(QueryDOI.doi == slug).first()
        if resolved:
            saved_search_type = 'doi'
            if resolved.requested_version:
                requested_version = resolved.requested_version
            else:
                # this should only be applicable for multisearch DOIs created before VDS
                # v6, which allowed different versions per resource; this rounds them
                # all up to  the newest version in the list. See #187 for more detail.
                requested_version = max(resolved.get_rounded_versions())
            result.update(
                {
                    'query': resolved.query,
                    'query_version': resolved.query_version,
                    'version': requested_version,
                    'resource_ids': resolved.get_resource_ids(),
                    'created': resolved.timestamp.isoformat(),
                }
            )

    if saved_search_type is None:
        # if both slug and DOI have failed
        raise toolkit.ValidationError('This saved search could not be found.')

    if result.get('query_version') == 'v0':
        result['query'] = convert_to_multisearch(result['query'])
        result['query_version'] = get_latest_query_version()

    # warn that "all resources" changes
    if len(result['resource_ids']) == 0:
        result['warnings'].append(
            toolkit._(
                'Resource IDs were not saved for this search. The list of available '
                'resources may have changed since this slug was saved.'
            )
        )

    # check that all the resources exist and are available
    try:
        valid_resource_ids = validate_datastore_resource_ids(result['resource_ids'])
    except toolkit.Invalid:
        valid_resource_ids = []
    valid_resource_count = len(valid_resource_ids)
    current_resource_count = len(result['resource_ids'])
    if valid_resource_count != current_resource_count:
        if valid_resource_count == 0:
            # this would change the search too much
            raise toolkit.ValidationError(
                'All resources associated with this search have been '
                'deleted, moved, or are no longer available.'
            )
        result['warnings'].append(
            toolkit._(
                'Some resources have been deleted, moved, or are no longer available. '
                'Affected resources: '
            )
            + str(current_resource_count - valid_resource_count)
        )
        if saved_search_type == 'doi' and hasattr(resolved, 'count'):
            result['warnings'].append(
                toolkit._('Record count at save time: ') + str(resolved.count)
            )
    result['resource_ids'] = valid_resource_ids

    return result


@action(schema.vds_slug_reserve(), helptext.vds_slug_reserve)
def vds_slug_reserve(context: dict, current_slug: str, new_reserved_slug: str):
    """
    Attempts to reserve the given slug under a new name. Each slug can only have one
    reserved name so if a reserved name already exists, the new name will overwrite it.

    Note that only sysadmins can update a reserved slug name, but normal users can add a
    reserved name.

    :param context: the CKAN action context
    :param current_slug: the current slug name
    :param new_reserved_slug: the new slug name
    :returns: the slug details as a dict
    """
    slug = resolve_slug(current_slug)
    if slug is None:
        raise toolkit.ValidationError(f'The slug {current_slug} does not exist')
    if slug.reserved_pretty_slug and not context['auth_user_obj'].sysadmin:
        raise toolkit.NotAuthorized(
            'Only sysadmins can replace existing reserved slugs.'
        )
    slug.reserved_pretty_slug = new_reserved_slug.lower()
    slug.commit()
    return slug.as_dict()



from ckanext.versioned_datastore.logic.multi.schema import multi_params
from ckanext.versioned_datastore.logic.validators import (
    boolean_validator,
    ignore_missing,
    not_empty,
    not_missing,
    url_safe,
)


def vds_slug_create() -> dict:
    return {
        **multi_params,
        'pretty_slug': [ignore_missing, boolean_validator],
        'nav_slug': [ignore_missing, boolean_validator],
    }


def vds_slug_resolve() -> dict:
    return {
        'slug': [str],
    }


def vds_slug_reserve() -> dict:
    return {
        'current_slug': [not_missing, not_empty, str],
        'new_reserved_slug': [not_missing, not_empty, str, url_safe],
    }



from ckan.plugins import toolkit
from ckantools.validators.ivalidators import BaseArgs

from ckanext.datastore.logic.schema import json_validator
from ckanext.versioned_datastore.lib.query.search.query import SchemaQuery
from ckanext.versioned_datastore.lib.query.utils import convert_to_multisearch
from ckanext.versioned_datastore.lib.utils import get_available_datastore_resources
from ckanext.versioned_datastore.logic.validators import (
    boolean_validator,
    ignore_missing,
    int_validator,
    not_missing,
    validate_datastore_resource_ids,
    validate_resource_ids,
)


class QueryArgs(BaseArgs):
    resource_ids: list
    query: dict
    query_version: str
    version: int
    slug_or_doi: str

    fields = {
        'resource_ids': [ignore_missing, validate_resource_ids],
        'query': [ignore_missing, json_validator],
        'query_version': [ignore_missing, str],
        'version': [ignore_missing, int_validator],
        'slug_or_doi': [ignore_missing, str],
    }

    def validate(self):
        if self.query and 'resource_ids' in self.query:
            validate_datastore_resource_ids(self.query['resource_ids'])
            if not self.resource_ids:
                self.resource_ids = self.query['resource_ids']

    def to_schema_query(self) -> SchemaQuery:
        query = self.query
        query_version = self.query_version
        resource_ids = self.resource_ids
        version = self.version

        if self.slug_or_doi:
            try:
                saved_query = toolkit.get_action('vds_slug_resolve')(
                    {}, {'slug': self.slug_or_doi}
                )
                query = saved_query.get('query')
                query_version = saved_query.get('query_version')
                resource_ids = saved_query.get('resource_ids')
                version = saved_query.get('version')
            except toolkit.ValidationError:
                # if the slug doesn't resolve, continue as normal
                pass

        if query_version and query_version.lower().startswith('v0'):
            # this is an old/basic query, so we need to convert it first
            query = convert_to_multisearch(query)
            query_version = None

        # if no resource IDs have been provided, use all resources available to the user
        if not resource_ids:
            resource_ids = list(get_available_datastore_resources())

        # schema query init will handle defaulting the various parts
        return SchemaQuery(resource_ids, version, query, query_version)


class DerivativeArgs(BaseArgs):
    format: str
    format_args: dict
    separate_files: bool
    ignore_empty_fields: bool
    transform: dict

    fields = {
        'format': [not_missing, str],
        'format_args': [ignore_missing, json_validator],
        'separate_files': [ignore_missing, boolean_validator],
        'ignore_empty_fields': [ignore_missing, boolean_validator],
        'transform': [ignore_missing, json_validator],
    }

    defaults = {
        'format_args': {},
        'separate_files': False,
        'ignore_empty_fields': False,
        'transform': {},
    }

    def ensure_defaults_are_set(self):
        for field, default_value in self.defaults.items():
            if getattr(self, field) is None:
                setattr(self, field, default_value)


class ServerArgs(BaseArgs):
    type: str
    type_args: dict
    custom_filename: str

    fields = {
        'type': [ignore_missing, str],
        'type_args': [ignore_missing, json_validator],
        'custom_filename': [ignore_missing, str],
    }

    defaults = {'type': 'direct', 'type_args': {}}


class NotifierArgs(BaseArgs):
    type: str
    type_args: dict

    fields = {
        'type': [ignore_missing, str],
        'type_args': [ignore_missing, json_validator],
    }

    defaults = {'type': 'none', 'type_args': {}}



"""
Remove version not null constraint.

Revision ID: 5932a36b7cf3
Revises: d2ca5da0573f
Create Date: 2024-09-30 18:55:43.328542
"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '5932a36b7cf3'
down_revision = 'd2ca5da0573f'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('versioned_datastore_import_stats', 'version', nullable=True)


def downgrade():
    op.alter_column('versioned_datastore_import_stats', 'version', nullable=False)



from ckantools.validators import object_validator

from ckanext.versioned_datastore.logic.download.arg_objects import (
    DerivativeArgs,
    NotifierArgs,
    QueryArgs,
    ServerArgs,
)
from ckanext.versioned_datastore.logic.validators import ignore_missing, not_missing


def vds_download_queue() -> dict:
    return {
        'query': [not_missing, object_validator(QueryArgs)],
        'file': [
            not_missing,
            object_validator(DerivativeArgs),
        ],
        # called 'file' instead of derivative to make its purpose clearer to the end user
        'server': [ignore_missing, object_validator(ServerArgs)],
        'notifier': [ignore_missing, object_validator(NotifierArgs)],
    }


def vds_download_regenerate():
    return {
        'download_id': [not_missing, str],
        'server': [ignore_missing, object_validator(ServerArgs)],
        'notifier': [ignore_missing, object_validator(NotifierArgs)],
    }



from typing import Optional

from ckan.logic import ValidationError
from ckan.plugins import toolkit
from ckantools.decorators import action
from jsonschema.exceptions import ValidationError

from ckanext.versioned_datastore.lib.query.schema import (
    get_latest_query_version,
    validate_query,
)
from ckanext.versioned_datastore.logic.schema import helptext, schema


@action(schema.vds_schema_latest(), helptext.vds_schema_latest, get=True)
def vds_schema_latest():
    """
    Retrieves the latest query schema version and returns it.

    :returns: the query schema version
    """
    return get_latest_query_version()


@action(schema.vds_schema_validate(), helptext.vds_schema_validate, get=True)
def vds_schema_validate(query: dict, query_version: Optional[str] = None):
    """
    Validates the given query against the given query schema version. If the query
    version is not provided, the latest query schema version is used.

    :param query: the query to validate
    :param query_version: the query schema version to validate against (default is None
        which means use the latest query schema version)
    :returns: True if the schema is valid, otherwise raises a CKAN ValidationError
    """
    if query_version is None:
        query_version = get_latest_query_version()
    try:
        validate_query(query, query_version)
    except ValidationError as e:
        raise toolkit.ValidationError(e.message)
    return True



from ckantools.decorators import auth


@auth(anon=True)
def vds_slug_create(context, data_dict) -> dict:
    return {'success': True}


@auth(anon=True)
def vds_slug_resolve(context, data_dict) -> dict:
    return {'success': True}


@auth(anon=True)
def vds_slug_edit(context, data_dict) -> dict:
    return {'success': True}



vds_multi_query = ''
vds_multi_count = ''
vds_multi_autocomplete_value = ''
vds_multi_autocomplete_field_latest = ''
vds_multi_autocomplete_field = ''
vds_multi_hash = ''
vds_multi_fields = ''
vds_multi_stats = ''
vds_multi_direct = ''
